import {
	Body,
	Controller,
	Get,
	Param,
	ParseIntPipe,
	Post,
	Request,
	StreamableFile
} from '@nestjs/common';
import { Responses } from 'dto';
import FileSystemService from '../services/filesystem';
import { UserRole } from '../entities';
import { Role } from '../authguards';

@Controller('api/fs')
export default class FileSystemController {
	constructor(private fsService: FileSystemService) {}

	@Get('root')
	@Role(UserRole.USER)
	async getRoot(@Request() req): Promise<Responses.FS.GetRootResponse> {
		return {
			statusCode: 200,
			rootId: req.user.rootId
		};
	}

	@Get('node/:node')
	@Role(UserRole.USER)
	async getNode(
		@Request() req,
		@Param('node', ParseIntPipe) nodeId
	): Promise<Responses.FS.GetNodeResponse> {
		const node = await this.fsService.getNodeAndValidate(nodeId, req.user);
		const data: Responses.FS.GetNodeResponse = {
			id: nodeId,
			statusCode: 200,
			name: node.name,
			parent: node.parentId,
			isFile: node.isFile
		};
		if (data.isFile) {
			data.size = node.size;
		} else {
			data.children = (await node.children).map((child) => child.id);
		}
		return data;
	}

	@Get('path/:node')
	@Role(UserRole.USER)
	async getPath(
		@Request() req,
		@Param('node', ParseIntPipe) nodeId
	): Promise<Responses.FS.GetPathResponse> {
		return {
			statusCode: 200,
			path: await this.fsService.generatePath(
				await this.fsService.getNodeAndValidate(nodeId, req.user)
			)
		};
	}

	@Post('createFolder')
	@Role(UserRole.USER)
	async createFolder(
		@Request() req,
		@Body('parent', ParseIntPipe) parent,
		@Body('name') name
	): Promise<Responses.FS.CreateFolderResponse> {
		const newChild = await this.fsService.create(
			await this.fsService.getNodeAndValidate(parent, req.user),
			name,
			req.user,
			false
		);
		return {
			statusCode: 200,
			id: newChild.id
		};
	}

	@Post('createFile')
	@Role(UserRole.USER)
	async createFile(
		@Request() req,
		@Body('parent', ParseIntPipe) parent,
		@Body('name') name
	): Promise<Responses.FS.CreateFileResponse> {
		const newChild = await this.fsService.create(
			await this.fsService.getNodeAndValidate(parent, req.user),
			name,
			req.user,
			true
		);
		return {
			statusCode: 200,
			id: newChild.id
		};
	}

	@Post('delete')
	@Role(UserRole.USER)
	async delete(
		@Request() req,
		@Body('node', ParseIntPipe) node_id
	): Promise<Responses.FS.DeleteResponse> {
		await this.fsService.delete(
			await this.fsService.getNodeAndValidate(node_id, req.user)
		);
		return { statusCode: 200 };
	}

	@Post('upload/:node')
	@Role(UserRole.USER)
	async upload(
		@Request() req,
		@Param('node', ParseIntPipe) nodeId
	): Promise<Responses.FS.UploadFileResponse> {
		await this.fsService.uploadFile(await req.file(), nodeId, req.user);
		return { statusCode: 200 };
	}

	@Post('download')
	@Role(UserRole.USER)
	async download(
		@Request() req,
		@Body('id', ParseIntPipe) id
	): Promise<StreamableFile> {
		return this.fsService.downloadFile(id, req.user);
	}
}
